import concurrent.futures
import os


def qimap(func, *iterables):
    with concurrent.futures.ThreadPoolExecutor(
        max_workers=int(os.getenv("QY_MAX_WORKERS",
                        min(32, os.cpu_count() + 4)))
    ) as executor:
        yield from executor.map(func, *iterables)


def qmap(func, *iterables):
    return list(qimap(func, *iterables))


def join(separator: str, *iterables: list[str]) -> str:
    return separator.join(iterables)



# coding: utf-8

"""
Created on 2024-08-15

"""
import code
from qy import *

LOCALS = globals()


def main():
    code.interact(
        banner='Qy Lang, a Lisp language implemented in and based on Python',
        readfunc=lambda prompt='': f"qy.eval((print, {input(prompt)}))",
        local=LOCALS,
        exitmsg='Goodbye!')



from pathlib import Path
import sys
import unittest
import lark
GRAMMER = f"""
?start: expressions
expressions: expression*

?expression: atom
    | list
    | quote_expression -> quote

?quote_expression: "'" expression
?list: "(" expression* ")"
?atom: STRING   -> string
    | SYMBOL    -> symbol
STRING: /"[^"]*"/
SYMBOL: /[a-zA-Z0-9_\-@:\.]+/
%import common.WS
%ignore WS
"""


class TestReader(unittest.TestCase):
    def setUp(self):
        sys.path.insert(0, '')

    def test_(self):
        from qy._reader import tokenize
        from qy import symbol
        tokens = tokenize(Path('examples/data.qy').read_text('utf-8'))
        print(list(tokens))
        parser = lark.Lark(GRAMMER)
        tree = parser.parse(Path('examples/data.qy').read_text('utf-8'))

        @lark.v_args(inline=True)
        class QyTransformer(lark.Transformer):
            def expressions(self, *tokens):
                return list(tokens)

            def quote(self, tokens):
                return ('quote', tokens)

            def string(self, token):
                return str(token[1:-1])

            def symbol(self, name):
                return symbol(name)

            def list(self, *items):
                return items
        transformer = QyTransformer()
        expr = transformer.transform(tree)
        print(expr)

    def test_reader(self):
        from qy.core import reader
        from qy import qy
        exp = reader("(+ 1 3)")[0]
        self.assertEqual(qy.eval(exp), 4)

    def tearDown(self):
        pass


if __name__ == '__main__':
    unittest.main()



import unittest
import sys
import asyncio


class TestOperator(unittest.TestCase):
    def setUp(self):
        sys.path.insert(0, '')

    def test_quote(self):
        from qy.operator import qy
        from qy.operator import quote
        self.assertEqual(qy.eval((quote, 'exp')), 'exp')
        self.assertEqual(qy.eval((quote, ('exp', 'exp'))), ('exp', 'exp'))

    def test_atom(self):
        from qy.operator import qy, T, NIL
        from qy.operator import atom, quote
        self.assertEqual(qy.eval((atom, 'exp')), T)
        self.assertEqual(qy.eval((atom, (quote, ('exp', 'exp')))), NIL)
        self.assertEqual(qy.eval((atom, (quote, ()))), T)

    def test_eq(self):
        from qy.operator import qy, T, NIL
        from qy.operator import eq, quote
        self.assertEqual(qy.eval((eq, 'exp', 'exp')), T)
        self.assertEqual(qy.eval((eq, 'exp', 'ex')), NIL)
        self.assertEqual(qy.eval((eq, (quote, ()), (quote, ()))), T)
        self.assertEqual(
            qy.eval((eq, (quote, ('exp',)), (quote, ('exp',)))), NIL)

    def test_car_cdr(self):
        from qy.operator import qy
        from qy.operator import car, cdr, quote
        self.assertEqual(qy.eval((car, (quote, ('exp', 'exp')))), 'exp')
        self.assertEqual(qy.eval((cdr, (quote, ('exp', 'exp')))), ('exp',))

    def test_cons(self):
        from qy.operator import qy
        from qy.operator import cons, quote
        self.assertEqual(
            qy.eval((cons, 'exp', (quote, ('exp',)))), ('exp', 'exp'))
        # (a b c)
        self.assertEqual(
            qy.eval((cons, 'a', (cons, 'b', (cons, 'c', (quote, ()))))),
            ('a', 'b', 'c')
        )

    def test_cond(self):
        from qy.operator import qy, T, NIL
        from qy.operator import cond, quote
        self.assertEqual(
            qy.eval(
                (cond,
                 (NIL, (quote, 'exp')),
                 (T, (quote, 'exp')))
            ),
            'exp'
        )
        self.assertEqual(
            qy.eval(
                (cond,
                 (NIL, (quote, 'exp')),
                 (NIL, (quote, 'exp'))
                 )
            ),
            NIL
        )

    def test_aeval(self):
        from qy import qy

        async def add(a, b):
            return a + b
        v = asyncio.run(qy.aeval((add, 1, 2)))
        self.assertEqual(v, 3)

    def tearDown(self):
        pass


if __name__ == '__main__':
    unittest.main()



# coding: utf-8

__version__ = '0.0.4'
__author__ = 'Ge'
__all__ = [
    'qy',
    'symbol',
    'symbolproxy',
    't', 'T',
    'nil', 'NIL',
    'quote', 'atom', 'eq', 'car', 'cdr', 'cons', 'cond'
]

"""
Created on 2024-08-15

"""

from qy.core import symbol, symbolproxy
from qy.operator import qy, T, NIL
from qy.operator import quote, atom, eq, car, cdr, cons, cond

t = T
nil = NIL



import unittest


class TestQyMulti(unittest.TestCase):
    def setUp(self):
        import sys
        sys.path.insert(0, '')

    def test_imap(self):
        from qy.operators import qimap
        self.assertEqual(
            list(qimap(lambda x: x + 1, [1, 2, 3])),
            [2, 3, 4]
        )
        self.assertEqual(
            list(qimap(lambda x, y: x + y, (1, 2, 3), (1, 2, 3))),
            [2, 4, 6]
        )

    def tearDown(self):
        pass


if __name__ == '__main__':
    unittest.main()



from typing import Literal
from qy.core import qy, symbol

T = qy.symbol('T', True)
NIL = qy.symbol('NIL', False)


@qy.operator('quote')
def quote(exp):
    return exp


@qy.operator('atom')
def atom(exp):
    if isinstance(exp, tuple) and len(exp) != 0:
        return NIL
    return T


@qy.operator('eq')
def eq(x, y) -> Literal['T', 'NIL']:
    # TODO: fix bug: the op may be not work correctly
    if x and y and isinstance(x, tuple) and isinstance(y, tuple):
        return NIL
    if not x and not y and isinstance(x, tuple) and isinstance(y, tuple):
        return T
    return T if x == y else NIL


@qy.operator('car')
def car(exp: tuple):
    return exp[0]


@qy.operator('cdr')
def cdr(exp: tuple):
    return exp[1:]


@qy.operator('cons')
def cons(x, y: tuple):
    return (x,) + y


@qy.operator('cond')
def cond(*cond_ressults: tuple):
    for c, r in cond_ressults:
        if not qy.eval(c) in [(), NIL, False]:
            return qy.eval(r)
    return NIL


@qy.operator('kwargs')
def kw(*args):
    return {args[i]: args[i + 1] for i in range(0, len(args), 2)}


@qy.operator('+')
def add(*args):
    return sum(args)


@qy.operator('-')
def sub(*args):
    return args[0] - sum(args[1:])


@qy.operator('*')
def mul(*args):
    result = 1
    for arg in args:
        result *= arg
    return result


@qy.operator('/')
def div(*args):
    result = args[0]
    for arg in args[1:]:
        result /= arg
    return result




from typing import Iterator, List
import unicodedata


class Token:
    def __init__(self, type: str, value: str) -> None:
        self.type = type
        self.value = value

    def __repr__(self) -> str:
        return f'<{self.type} "{self.value}">'


def tokenize(source: str) -> Iterator[Token]:
    source = unicodedata.normalize('NFKC', source)
    buffer: List[str] = []
    in_string = False
    i = 0

    while i < len(source):
        char = source[i]

        if in_string:
            if char == '"' and source[i-1] != '\\':
                buffer.append(char)
                yield Token('string', ''.join(buffer))
                buffer.clear()
                in_string = False
            else:
                buffer.append(char)
        else:
            if char.isspace():
                if buffer:
                    yield Token('atom', ''.join(buffer))
                    buffer.clear()
                yield Token('whitespace', char)
            elif char in '()':
                if buffer:
                    yield Token('atom', ''.join(buffer))
                    buffer.clear()
                yield Token('paren', char)
            elif char == '"':
                if buffer:
                    yield Token('atom', ''.join(buffer))
                    buffer.clear()
                buffer.append(char)
                in_string = True
            elif char == "'":
                if buffer:
                    yield Token('atom', ''.join(buffer))
                    buffer.clear()
                yield Token('quote', char)
            elif char == ';':
                if buffer:
                    yield Token('atom', ''.join(buffer))
                    buffer.clear()
                comment_buffer = [char]
                i += 1
                while i < len(source) and source[i] != '\n':
                    comment_buffer.append(source[i])
                    i += 1
                if i < len(source):  # include the newline in the comment
                    comment_buffer.append(source[i])
                yield Token('comment', ''.join(comment_buffer))
            else:
                buffer.append(char)

        i += 1

    if buffer:
        if in_string:
            yield Token('string', ''.join(buffer))
        else:
            yield Token('atom', ''.join(buffer))


def untokenize(tokens: List[Token]) -> str:
    result = []
    indent_level = 0
    need_space = False

    for i, token in enumerate(tokens):
        if token.type == 'whitespace':
            if '\n' in token.value:
                result.append('\n' + '  ' * indent_level)
                need_space = False
            elif need_space:
                result.append(' ')
                need_space = False
        elif token.type == 'comment':
            result.append(token.value)
            need_space = False
        elif token.type == 'string':
            if need_space:
                result.append(' ')
            result.append(token.value)
            need_space = True
        elif token.type == 'paren':
            if token.value == '(':
                if need_space:
                    result.append(' ')
                result.append(token.value)
                indent_level += 1
                need_space = False
            elif token.value == ')':
                indent_level = max(0, indent_level - 1)
                result.append(token.value)
                need_space = True
        elif token.type == 'quote':
            if need_space:
                result.append(' ')
            result.append(token.value)
            need_space = False
        elif token.type == 'atom':
            if need_space:
                result.append(' ')
            result.append(token.value)
            need_space = True

        # Handle special cases for improved formatting
        if i < len(tokens) - 1:
            next_token = tokens[i+1]
            if token.type == 'paren' and token.value == '(' and next_token.type == 'paren' and next_token.value == ')':
                need_space = False
            elif token.type == 'quote' and next_token.type != 'whitespace':
                need_space = False

    return ''.join(result).strip()



# coding: utf-8

__all__ = [
    'symbol',
    'symbolproxy',
    'qy'
]

"""
Created on 2024-08-15

"""

import traceback
from typing import Callable, Any, Union
import warnings

import lark
PYOBJECT = object
ATOM = Union[
    str, int, float, bytes, bool, None, 'symbol',
    object
]

SEXPRESSION = Union[ATOM, tuple[ATOM]]

INTERMEDIATE_LANG = tuple[
    'INTERMEDIATE_LANG',
    'symbol',
    int, float, str, bool, None, bytes
]

GRAMMER = f"""
?start: expressions
expressions: expression*
?expression: atom
    | list
    | quote_expression -> quote
?quote_expression: "'" expression
?list: "(" expression* ")"
?atom: STRING   -> string
    | SYMBOL    -> symbol
STRING: /"[^"]*"/
SYMBOL: /[a-zA-Z0-9_\-@:\.\+\/]+/
%import common.WS
%ignore WS
"""

parser = lark.Lark(GRAMMER)


@lark.v_args(inline=True)
class QyTransformer(lark.Transformer):
    def expressions(self, *tokens):
        return list(tokens)

    def quote(self, tokens):
        return ('quote', tokens)

    def string(self, token):
        return str(token[1:-1])

    def symbol(self, name):
        return symbolproxy(name)

    def list(self, *items):
        return items


def reader(source: str):
    tree = parser.parse(source)
    return QyTransformer().transform(tree)


class QyError(Exception):
    ...


class QySyntaxError(QyError):
    ...


class QyRuntimeError(QyError):
    ...


class QySymbolError(QyError):
    ...


class QyEvelError(QyError):
    pass


class symbol:
    """
    symbol is different from other symbol.
    we can use symbol as a function, and we can use symbol as a value.
    it should be differenciated from the symbol in the symbol space.

    when the symbol is used as a function, it should be evaluated.
    when the symbol is used as a value, it should be returned

    so, if the symbol is the first element of the s-expression, 
        evaluator will use symbol.__call__ to evaluate it.
        for operator, user can require evaluator use orginal value (s-exp)
        ...
    if the symbol is not the first element,
        evaluator will use symbol.value to return the symbol object.

    name is the name of the symbol.

    """
    __solts__ = ('name', 'value', 'require_eval')

    def __init__(self, name: str, value: Any = None) -> None:
        self.name = name
        self.value = value

    def __eq__(self, value: object) -> bool:
        return (
            isinstance(value, symbol)
            and value.name == self.name
            and value.value == self.value
        )

    def __hash__(self) -> int:
        return hash((self.name, self.value))

    def __repr__(self) -> str:
        return f'<symbol {self.name}>'

    def __str__(self) -> str:
        return self.name

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        return self.value(*args, **kwds)

    def __bool__(self) -> bool:
        return bool(self.value)


class symbolproxy:
    __solts__ = ('name')

    def __init__(self, name: str) -> None:
        self.name = name

    def __repr__(self) -> str:
        return f'<symbolproxy {self.name}>'


class Qy:

    def __init__(self, *, thread=False) -> None:
        self.SYMBOLSPACE: dict[str, symbol] = {}

    def symbol(self, name: str, value: Any = None) -> symbol:
        s = symbol(name, value)
        if name in self.SYMBOLSPACE:  # warning
            warnings.warn(f'{name} is already in the symbol space')
        self.SYMBOLSPACE[name] = s
        return s

    def operator(self, name: str, func: Callable = None) -> None:
        if not isinstance(name, str):
            raise TypeError('name must be str')
        if func is None:
            return lambda func: self.operator(name, func)
        if not callable(func):
            raise TypeError('func must be callable')
        s = symbol(name, func)
        self.SYMBOLSPACE[name] = s
        return s

    def eval(self, s_expression: SEXPRESSION):
        """
        if s-expression is atom (symbol? NIL T), return it.

        if s-exp only is exp(need eval) or aotm

        TODO: should I return the symbol object or the value of the symbol?

        """
        # symbol evaluation
        if isinstance(s_expression, symbol):
            return s_expression.value
        if isinstance(s_expression, symbolproxy):
            if s_expression.name in self.SYMBOLSPACE:
                return self.SYMBOLSPACE[s_expression.name].value
            try:
                return int(s_expression.name)
            except ValueError:
                pass
            try:
                return float(s_expression.name)
            except ValueError:
                pass
            raise QySymbolError(
                f'{s_expression.name} is not in the symbol space')
        if not isinstance(s_expression, tuple):
            return s_expression

        # s-expression evaluation
        operator, *arguments = s_expression

        # search symbol from symbol space
        if isinstance(operator, symbolproxy):
            operator: symbol = self.SYMBOLSPACE[operator.name]

        # TODO: optimize the code
        if isinstance(operator, symbol):
            from .operator import quote, car, cdr, cons, cond
            if operator is quote:
                if len(arguments) != 1:
                    raise QyEvelError('Error: quote')
                return arguments[0]
            if operator is car:
                if len(arguments) != 1:
                    raise QyEvelError('Error: car')
                return car(self.eval(arguments[0]))
            if operator is cdr:
                if len(arguments) != 1:
                    raise QyEvelError('Error: cdr')
                return cdr(self.eval(arguments[0]))
            if operator is cons:
                if len(arguments) != 2:
                    raise QyEvelError('Error: cons')
                return cons(self.eval(arguments[0]), self.eval(arguments[1]))
            if operator is cond:
                return cond(*arguments)
        try:
            from .operator import kw
            args, kwargs = [], {}
            for arg in arguments:
                if isinstance(arg, tuple) and arg and arg[0] is kw:
                    kwargs.update(kw(*arg[1:]))
                else:
                    args.append(arg)
            for k, v in kwargs.items():
                kwargs[k] = self.eval(v)
            return operator(*map(self.eval, args), **kwargs)
        except SystemExit as e:
            raise e from None
        except BaseException as e:
            e = '\n'.join(traceback.format_exception(e))
            raise QyEvelError(
                f'Error: {operator} {tuple(arguments)}\n\n{e}'
            ) from None

    async def aeval(self, s_expression: SEXPRESSION):
        from .operator import T, NIL, atom

        if atom(s_expression) is T:
            return s_expression

        # when s-expression is (), return NIL
        # if you want to return (), you should use quote
        if s_expression == ():
            return NIL

        if not isinstance(s_expression, tuple):
            if isinstance(s_expression, (symbol, symbolproxy)):
                return s_expression.value
            return s_expression

        operator, *arguments = s_expression

        if isinstance(operator, (symbol, symbolproxy)):
            from .operator import quote, car, cdr, cons, cond
            if operator is quote:
                if len(arguments) != 1:
                    raise QyEvelError('Error: quote')
                return arguments[0]
            if operator is car:
                if len(arguments) != 1:
                    raise QyEvelError('Error: car')
                return car(await self.aeval(arguments[0]))
            if operator is cdr:
                if len(arguments) != 1:
                    raise QyEvelError('Error: cdr')
                return cdr(await self.aeval(arguments[0]))
            if operator is cons:
                if len(arguments) != 2:
                    raise QyEvelError('Error: cons')
                return cons(await self.aeval(arguments[0]), await self.aeval(arguments[1]))
            if operator is cond:
                return await cond(*arguments)
        try:
            arguments_result = []

            for arg in arguments:
                arguments_result.append(await self.aeval(arg))
            return await operator(*arguments_result)
        except SystemExit as e:
            raise e
        except BaseException as e:
            e = '\n'.join(traceback.format_exception(e))
            raise QyEvelError(
                f'Error: {operator} {arguments}\n\n{e}'
            ) from None

    def exec(self, ast: INTERMEDIATE_LANG):
        self.eval(ast)


qy = Qy()



from setuptools import setup, find_packages
from pathlib import Path
import re
from pathlib import Path


def get_version():
    code = Path('qy/__init__.py').read_text(encoding='utf-8')
    return re.search(r"^__version__ = '([^']+)'", code, re.M).group(1)


setup(
    name='QyLang',
    version=get_version(),
    description='Qy Lang, a Lisp language implemented in and based on Python',
    long_description=Path('README.md').read_text(encoding='utf-8'),
    long_description_content_type='text/markdown',
    author='Ge',
    author_email='cosplox@outlook.com',
    url='https://github.com/Cosmic-Developers-Union/Qy',
    install_requires=[
        'lark>=1.2.2'
    ],
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Lisp',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.8',
    entry_points={
        'console_scripts': [
            'qy=qy.cli:main',  # Assuming you have a CLI entry point
        ],
    },
    license='Apache License 2.0'
)
